from enum import Enum
from functools import partial
from typing import Awaitable, Callable, Dict, Iterable, Optional, Union

import arrow
from fastapi import APIRouter, BackgroundTasks, Depends, Header, Request, Response
from fastapi import status as fastapi_status
from fastapi.encoders import jsonable_encoder
from fastapi.params import Depends as DependsType
from fastapi.responses import JSONResponse
from globus_action_provider_tools import (
    ActionProviderDescription,
    ActionRequest,
    ActionStatus,
    ActionStatusValue,
    AuthState,
    TokenChecker,
)
from globus_action_provider_tools.authorization import (
    authorize_action_access_or_404,
    authorize_action_management_or_404,
)
from globus_action_provider_tools.errors import AuthenticationError
from globus_sdk import ConfidentialAppAuthClient
from pydantic import create_model

from .async_helpers import async_invoke
from .fastapi_action_provider_exceptions import (
    ActionConflictException,
    ActionMethodMissingImplementation,
    ActionNotFoundException,
    ApplicationError,
    UnauthorizedException,
)
from .persistence import ActionProviderPersistence, ActionProviderPersistenceReturnType

StatusInOutCallable = Callable[
    [ActionStatus], Union[ActionStatus, Awaitable[ActionStatus]]
]

_EMPTY_STATUS_DETAILS = "EMPTY STATUS MUST BE FILLED IN"


class ImplementationBehavior(Enum):
    REQUIRED = 1
    REQUIRED_WITHOUT_BACKEND = 2
    NOT_REQUIRED = 3
    UNKNOWN_METHOD = 4


_api_method_required_policies = {
    "run": ImplementationBehavior.REQUIRED,
    "status": ImplementationBehavior.REQUIRED_WITHOUT_BACKEND,
    "cancel": ImplementationBehavior.NOT_REQUIRED,
    "release": ImplementationBehavior.REQUIRED_WITHOUT_BACKEND,
    "resume": ImplementationBehavior.NOT_REQUIRED,
    "log": ImplementationBehavior.NOT_REQUIRED,
}


def _authorize_access(
    checker_function: Callable[[ActionStatus, AuthState], None],
    action: Optional[ActionStatus],
    auth_state: AuthState,
    req_action_id: str = "<Unknown>",
    raise_on_missing_action=True,
    exception_class=ActionNotFoundException,
):
    if action is None and raise_on_missing_action:
        raise exception_class(req_action_id)
    if action is not None:
        try:
            checker_function(action, auth_state)
        except AuthenticationError:
            raise exception_class(action.action_id)


def authorize_action_access(
    action: Optional[ActionStatus],
    auth_state: AuthState,
    req_action_id: str = "<Unknown>",
    raise_on_missing_action=True,
    exception_class=ActionNotFoundException,
):
    _authorize_access(
        authorize_action_access_or_404,
        action,
        auth_state,
        req_action_id=req_action_id,
        raise_on_missing_action=raise_on_missing_action,
        exception_class=exception_class,
    )


def authorize_action_management(
    action: Optional[ActionStatus],
    auth_state: AuthState,
    req_action_id: str = "<Unknown>",
    raise_on_missing_action=True,
    exception_class=ActionNotFoundException,
):
    _authorize_access(
        authorize_action_management_or_404,
        action,
        auth_state,
        req_action_id=req_action_id,
        raise_on_missing_action=raise_on_missing_action,
        exception_class=exception_class,
    )


def fill_action_status(
    in_action: ActionStatus,
    req: Optional[ActionRequest] = None,
    auth_state: Optional[AuthState] = None,
) -> ActionStatus:
    if auth_state is not None:
        in_action.creator_id = auth_state.effective_identity

    if req is not None:
        if not in_action.monitor_by and req.monitor_by:
            in_action.monitor_by = req.monitor_by
        if not in_action.manage_by and req.manage_by:
            in_action.manage_by = req.manage_by
    if in_action.is_complete() and in_action.completion_time is None:
        in_action.completion_time = str(arrow.utcnow())
    if in_action.release_after is None:
        #        in_action.release_after = datetime.timedelta(days=30)
        in_action.release_after = "P30D"
    if in_action.display_status is None:
        in_action.display_status = str(in_action.status)

    return in_action


def missing_handler(
    route_name: str,
    backend_present: bool,
    *args,
    action_request: Optional[ActionRequest] = None,
    auth_state: Optional[AuthState] = None,
    **kwargs,
) -> Optional[ActionStatus]:
    behavior = _api_method_required_policies.get(
        route_name, ImplementationBehavior.UNKNOWN_METHOD
    )
    if behavior is ImplementationBehavior.REQUIRED or (
        behavior is ImplementationBehavior.REQUIRED_WITHOUT_BACKEND
        and backend_present is False
    ):
        raise ActionMethodMissingImplementation(route_name)
    # Since a common pattern is ActionStatus in the input list and an expected
    # ActionStatus return, we find it in the args list and return it if present
    for arg in args:
        if isinstance(arg, ActionStatus):
            return fill_action_status(arg, action_request, auth_state)
    return None


async def check_token(
    checker: TokenChecker, authorization_header: Optional[str] = None
) -> AuthState:

    token = ""
    if authorization_header is not None:
        authorization_header = authorization_header.strip()
        if authorization_header.startswith("Bearer "):
            token = authorization_header.lstrip("Bearer ")

    auth_state = checker.check_token(token)
    return auth_state


def fastapi_auth_dependency(
    token_checker: TokenChecker,
) -> DependsType:
    async def token_checker_dependency(
        authorization: Optional[str] = Header(None),
    ) -> Awaitable[AuthState]:
        return await check_token(token_checker, authorization)

    return Depends(token_checker_dependency)


class GlobusActionProviderRouter(APIRouter):
    def __init__(
        self,
        ap_description: ActionProviderDescription,
        globus_auth_client_id: str,
        globus_auth_client_secret: str,
        *args,
        persistence_backend: Optional[ActionProviderPersistence] = None,
        additional_scopes: Iterable[str] = (),
        **kwargs,
    ):
        self.auth_client = ConfidentialAppAuthClient(
            globus_auth_client_id, globus_auth_client_secret
        )
        self.ap_description = ap_description
        self.persistence_backend = persistence_backend
        self.handlers: Dict[str, StatusInOutCallable] = {
            k: partial(missing_handler, k, self.persistence_backend is not None)
            for k in _api_method_required_policies.keys()
        }

        scopes = [ap_description.globus_auth_scope]
        scopes.extend(additional_scopes)
        self.token_checker = self._create_token_checker(
            globus_auth_client_id, globus_auth_client_secret, scopes
        )

        super().__init__(*args, **kwargs)
        self.register_api_routes()

    def _create_token_checker(
        self,
        globus_auth_client_id: str,
        globus_auth_client_secret: str,
        scopes: Iterable[str],
    ):
        return TokenChecker(globus_auth_client_id, globus_auth_client_secret, scopes)

    def globus_auth_dependency(self) -> DependsType:
        return fastapi_auth_dependency(self.token_checker)

    def register_api_routes(self) -> None:
        ThisActionRequest = create_model(
            "ThisActionRequest",
            __base__=ActionRequest,
            body=(self.ap_description.input_schema, ...),
        )

        auth_state_dependency = self.globus_auth_dependency()

        @self.get("/")
        async def introspect(
            auth_state: AuthState = auth_state_dependency,
        ) -> ActionProviderDescription:
            return await self.introspect_handler(auth_state)

        @self.post("/run", status_code=fastapi_status.HTTP_202_ACCEPTED)
        async def run(
            req: ThisActionRequest,
            request: Request,
            response: Response,
            background_tasks: BackgroundTasks,
            auth_state: AuthState = auth_state_dependency,
        ) -> ActionStatus:
            status = await self.run_handler(
                req, auth_state, request, response, background_tasks
            )
            return status

        @self.get("/{action_id}/status")
        async def status(
            action_id: str,
            request: Request,
            response: Response,
            background_tasks: BackgroundTasks,
            auth_state: AuthState = auth_state_dependency,
        ) -> ActionStatus:
            return await self.status_handler(
                action_id, auth_state, request, response, background_tasks
            )

        @self.post("/{action_id}/resume")
        async def resume(
            action_id: str,
            request: Request,
            response: Response,
            background_tasks: BackgroundTasks,
            auth_state: AuthState = auth_state_dependency,
        ) -> ActionStatus:
            return await self.resume_handler(
                action_id, auth_state, request, response, background_tasks
            )

        @self.post("/{action_id}/cancel")
        async def cancel(
            action_id: str,
            request: Request,
            response: Response,
            background_tasks: BackgroundTasks,
            auth_state: AuthState = auth_state_dependency,
        ) -> ActionStatus:
            return await self.cancel_handler(
                action_id, auth_state, request, response, background_tasks
            )

        @self.post("/{action_id}/release")
        async def release(
            action_id: str,
            request: Request,
            response: Response,
            background_tasks: BackgroundTasks,
            auth_state: AuthState = auth_state_dependency,
        ) -> ActionStatus:
            return await self.release_handler(
                action_id, auth_state, request, response, background_tasks
            )

        @self.get("/{action_id}/log")
        async def log(
            action_id: str,
            request: Request,
            response: Response,
            background_tasks: BackgroundTasks,
            auth_state: AuthState = auth_state_dependency,
        ) -> ActionStatus:
            return await self.log_handler(
                action_id, auth_state, request, response, background_tasks
            )

    def create_stub_action_status(
        self, action_id: Optional[str], creator_id: Optional[str]
    ) -> ActionStatus:
        action_props = dict(
            status=ActionStatusValue.ACTIVE,
            creator_id=creator_id,
            details=_EMPTY_STATUS_DETAILS,
        )
        if action_id is not None:
            action_props["action_id"] = action_id
        return ActionStatus(**action_props)

    async def lookup_status(
        self, action_id: str, raise_on_not_found=True
    ) -> ActionProviderPersistenceReturnType:
        if self.persistence_backend is None:
            return (None, None, None)
        (
            status,
            request,
            extra_data,
        ) = await self.persistence_backend.lookup_by_action_id(action_id)
        if status is None:
            if raise_on_not_found:
                raise ActionNotFoundException(action_id)
            status = (self.create_stub_action_status(action_id, None),)
        return status, request, extra_data

    async def lookup_by_request_id(
        self, request_id: str, creator_id: str
    ) -> ActionProviderPersistenceReturnType:
        if self.persistence_backend is None:
            return (None, None, None)
        (
            status,
            request,
            extra_data,
        ) = await self.persistence_backend.lookup_by_request_id_and_identity(
            request_id, creator_id
        )
        if status is None:
            status = self.create_stub_action_status(None, creator_id)
        return status, request, extra_data

    async def introspect_handler(self, auth_state: AuthState) -> JSONResponse:
        if not auth_state.check_authorization(
            self.ap_description.visible_to, allow_public=True
        ):
            raise UnauthorizedException()

        def model_class_encoder(model_class) -> Dict:
            r = model_class.schema()
            return r

        json_vals = jsonable_encoder(
            self.ap_description,
            custom_encoder={
                type(self.ap_description.input_schema): model_class_encoder
            },
        )
        return JSONResponse(content=json_vals)

    async def run_handler(
        self,
        req: ActionRequest,
        auth_state: AuthState,
        request: Request,
        response: Response,
        background_tasks: BackgroundTasks,
    ) -> ActionStatus:
        if not auth_state.check_authorization(
            self.ap_description.runnable_by, allow_all_authenticated_users=True
        ):
            raise UnauthorizedException()
        status, saved_req, extra_data = await self.lookup_by_request_id(
            req.request_id, auth_state.effective_identity
        )
        if saved_req is not None:
            if saved_req.body != req.body.json():
                breakpoint()
                raise ActionConflictException(
                    saved_req.request_id,
                    (
                        f"Request with id {saved_req.request_id} with different body "
                        "payload already processed"
                    ),
                )
            elif status is not None:
                return await self.status_handler(
                    status.action_id, auth_state, request, response, background_tasks
                )
            else:
                raise ApplicationError(
                    "Unable to find action associated with request id "
                    f"{saved_req.request_id}"
                )
        if extra_data is None:
            extra_data = {}
        status = await async_invoke(
            self.handlers["run"],
            req,
            action_status=status,
            action_request=req,
            auth_state=auth_state,
            fastapi_request=request,
            fastapi_response=response,
            extra_data=extra_data,
            background_tasks=background_tasks,
        )
        status = fill_action_status(status, req, auth_state)
        if self.persistence_backend is not None and status is not None:
            await self.persistence_backend.store_action(
                status, req, auth_state.effective_identity, extra_data
            )
        return status

    async def status_handler(
        self,
        action_id: str,
        auth_state: AuthState,
        request: Request,
        response: Response,
        background_tasks: BackgroundTasks,
    ) -> ActionStatus:
        status, req, extra_data = await self.lookup_status(action_id)
        authorize_action_access(status, auth_state)
        if status is not None and status.is_complete():
            return status
        res = await async_invoke(
            self.handlers["status"],
            status,
            action_status=status,
            action_id=action_id,
            action_request=req,
            auth_state=auth_state,
            fastapi_request=request,
            fastapi_response=response,
            extra_data=extra_data,
            background_tasks=background_tasks,
        )
        res = fill_action_status(res, req, auth_state)
        if self.persistence_backend is not None and status is not None:
            await self.persistence_backend.store_action(
                status, req, auth_state.effective_identity, extra_data
            )
        return res

    async def release_handler(
        self,
        action_id: str,
        auth_state: AuthState,
        request: Request,
        response: Response,
        background_tasks: BackgroundTasks,
    ) -> ActionStatus:
        status, req, extra_data = await self.lookup_status(action_id)
        authorize_action_management(status, auth_state)
        if status is not None and not status.is_complete():
            raise ActionConflictException(
                action_id,
                detail="Cannot release an Action which is not in a completed state.",
            )
        res = await async_invoke(
            self.handlers["release"],
            status,
            action_status=status,
            action_id=action_id,
            action_request=req,
            auth_state=auth_state,
            fastapi_request=request,
            fastapi_response=response,
            extra_data=extra_data,
            background_tasks=background_tasks,
        )
        res = fill_action_status(res, req, auth_state)
        if self.persistence_backend is not None and status is not None:
            await self.persistence_backend.remove_action(action_id)
        return res

    async def cancel_handler(
        self,
        action_id: str,
        auth_state: AuthState,
        request: Request,
        response: Response,
        background_tasks: BackgroundTasks,
    ) -> ActionStatus:
        status, req, extra_data = await self.lookup_status(action_id)
        authorize_action_management(status, auth_state)
        if status is not None and status.is_complete():
            return status
        res = await async_invoke(
            self.handlers["cancel"],
            status,
            action_status=status,
            action_id=action_id,
            action_request=req,
            auth_state=auth_state,
            fastapi_request=request,
            fastapi_response=response,
            extra_data=extra_data,
            background_tasks=background_tasks,
        )
        res = fill_action_status(res, req, auth_state)
        if self.persistence_backend is not None and status is not None:
            await self.persistence_backend.store_action(
                status, req, auth_state.effective_identity, extra_data
            )
        return res

    async def log_handler(
        self,
        action_id: str,
        auth_state: AuthState,
        request: Request,
        response: Response,
        background_tasks: BackgroundTasks,
    ) -> ActionStatus:
        status, req, extra_data = await self.lookup_status(action_id)
        authorize_action_access(status, auth_state)
        # TODO TODO: This is really all wrong as log is totally different
        # TODO: Check if it is already complete so we don't need to invoke the handler
        res = await async_invoke(
            self.handlers["log"],
            status,
            action_status=status,
            action_id=action_id,
            action_request=req,
            auth_state=auth_state,
            fastapi_request=request,
            fastapi_response=response,
            extra_data=extra_data,
            background_tasks=background_tasks,
        )
        # TODO: Store the returned state in the backend
        return res

    async def resume_handler(
        self,
        action_id: str,
        auth_state: AuthState,
        request: Request,
        response: Response,
        background_tasks: BackgroundTasks,
    ) -> ActionStatus:
        status, req, extra_data = await self.lookup_status(action_id)
        authorize_action_management(status, auth_state)
        if status is not None and status.is_complete():
            return status
        res = await async_invoke(
            self.handlers["resume"],
            status,
            action_status=status,
            action_id=action_id,
            action_request=req,
            auth_state=auth_state,
            fastapi_request=request,
            fastapi_response=response,
            extra_data=extra_data,
            background_tasks=background_tasks,
        )
        res = fill_action_status(res, req, auth_state)
        if self.persistence_backend is not None and status is not None:
            await self.persistence_backend.store_action(
                status, req, auth_state.effective_identity, extra_data
            )
        return res

    def run(
        self, f: Callable[[ActionRequest], Union[ActionStatus, Awaitable[ActionStatus]]]
    ) -> Callable[[ActionRequest], ActionStatus]:
        self.handlers["run"] = f
        return f

    def status(self, f: StatusInOutCallable) -> Callable:
        self.handlers["status"] = f
        return f

    def resume(self, f: StatusInOutCallable) -> Callable:
        self.handlers["resume"] = f
        return f

    def cancel(self, f: StatusInOutCallable) -> Callable:
        self.handlers["cancel"] = f
        return f

    def release(self, f: StatusInOutCallable) -> Callable:
        self.handlers["release"] = f
        return f

    def log(self, f: StatusInOutCallable) -> Callable:
        self.handlers["log"] = f
        return f



from typing import Any, Dict, Optional, Tuple

from globus_action_provider_tools import ActionRequest, ActionStatus

from .persistence import ActionProviderPersistence, ActionProviderPersistenceReturnType


class InMemoryActionProviderPersistence(ActionProviderPersistence):
    def __init__(self):
        self.action_store: Dict[str, ActionProviderPersistenceReturnType] = {}
        self.by_request_id_index: Dict[Tuple[str, str], str] = {}

    async def lookup_by_action_id(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        return self.action_store.get(action_id, (None, None, None))

    async def lookup_by_request_id_and_identity(
        self, request_id: str, user_identity: str
    ) -> ActionProviderPersistenceReturnType:
        action_id = self.by_request_id_index.get((request_id, user_identity))
        if action_id is not None:
            return await self.lookup_by_action_id(action_id)
        else:
            return (None, None, None)

    async def store_action(
        self,
        action: ActionStatus,
        request: Optional[ActionRequest] = None,
        creator_id: Optional[str] = None,
        extra_data: Optional[Dict[str, Any]] = None,
    ) -> ActionProviderPersistenceReturnType:
        self.action_store[action.action_id] = (action, request, extra_data)
        if request is not None and creator_id is not None:
            self.by_request_id_index[
                (request.request_id, creator_id)
            ] = action.action_id
        return action, request, extra_data

    async def remove_action(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        action, request, extra_data = await self.lookup_by_action_id(action_id)
        if action is not None:
            self.action_store.pop(action_id, None)
            if request is not None:
                self.by_request_id_index.pop(
                    (request.request_id, action.creator_id), None
                )
        return action, request, extra_data



import uuid
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Tuple

from globus_action_provider_tools import (
    ActionProviderJsonEncoder,
    ActionRequest,
    ActionStatus,
)


class PersistenceJsonEncoder(ActionProviderJsonEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj)
        else:
            return super(PersistenceJsonEncoder, self).default(obj)


ActionProviderPersistenceReturnType = Tuple[
    Optional[ActionStatus], Optional[ActionRequest], Optional[Dict[str, Any]]
]


class ActionProviderPersistence(ABC):
    @abstractmethod
    async def lookup_by_action_id(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        ...

    @abstractmethod
    async def lookup_by_request_id_and_identity(
        self, request_id: str, user_identity: str
    ) -> ActionProviderPersistenceReturnType:
        ...

    @abstractmethod
    async def store_action(
        self,
        action: ActionStatus,
        request: Optional[ActionRequest] = None,
        creator_id: Optional[str] = None,
        extra_data: Optional[Dict[str, Any]] = None,
    ) -> ActionProviderPersistenceReturnType:
        ...

    @abstractmethod
    async def remove_action(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        ...



import json
import logging
from typing import Any, Dict, Optional, Type, TypeVar

import boto3
from boto3.dynamodb.conditions import Attr, Key
from botocore.exceptions import ClientError, EndpointConnectionError
from globus_action_provider_tools import ActionRequest, ActionStatus, ActionStatusValue
from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource
from mypy_boto3_dynamodb.service_resource import Table as DynamoTable
from pydantic import BaseModel

from .persistence import (
    ActionProviderPersistence,
    ActionProviderPersistenceReturnType,
    PersistenceJsonEncoder,
)

log = logging.getLogger(__name__)

_attribute_definitions = ({"AttributeName": "action_id", "AttributeType": "S"},)

_key_schema = ({"AttributeName": "action_id", "KeyType": "HASH"},)

T = TypeVar("T", bound=BaseModel)

_action_property_conversions = {
    "status": lambda x: ActionStatusValue[x],
}


def create_dynamo_client(**kwargs) -> DynamoDBServiceResource:
    client = boto3.resource("dynamodb", **kwargs)
    return client


def _json_to_object(json_val: Optional[str], model_class: Type[T]) -> Optional[T]:
    if json_val is None:
        return None
    props = json.loads(json_val)
    for val_to_convert, convert_fn in _action_property_conversions.items():
        dynamo_val = props.get(val_to_convert)
        if dynamo_val is not None:
            props[val_to_convert] = convert_fn(dynamo_val)
    return model_class(**props)


def _dynamo_item_to_return_vals(
    item: Optional[Dict[str, Any]]
) -> ActionProviderPersistenceReturnType:
    if item is None:
        return None, None, None
    action_status = _json_to_object(item.get("action_status"), ActionStatus)
    request = _json_to_object(item.get("action_request"), ActionRequest)
    extra_data = item.get("extra_data")
    return action_status, request, extra_data


class DynamoActionProviderPersistence(ActionProviderPersistence):
    def __init__(
        self,
        dynamo_client: DynamoDBServiceResource,
        table_name: str,
        create_table=False,
        create_table_args: Optional[Dict] = None,
    ):
        self.client = dynamo_client
        self.table_name = table_name
        if create_table:
            if create_table_args is None:
                create_table_args = {}
            self.table = self.create_table(self.table_name, **create_table_args)
        else:
            self.table = self.client.Table(table_name)

    def create_table(self, table_name: Optional[str] = None, **kwargs) -> DynamoTable:
        if table_name is None:
            table_name = self.table_name
        try:
            self.client.create_table(
                TableName=table_name,
                AttributeDefinitions=_attribute_definitions,
                KeySchema=_key_schema,
                **kwargs,
            )
        except ClientError as ce:
            if ce.response["Error"]["Code"] == "ResourceInUseException":
                log.info(f"Dynamo table {table_name} already exists.")
            else:
                log.info(f"Error creating dynamo table {table_name}: {ce}")
        except EndpointConnectionError as err:
            log.info(f"Error creating dynamo table {table_name}: {err}")

        return self.client.Table(table_name)

    async def lookup_by_action_id(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        response = self.table.query(
            KeyConditionExpression=Key("action_id").eq(action_id)
        )
        items = response.get("Items")

        if items is None:
            item = None
        elif len(items) == 0:
            item = None
        else:
            item = items[0]
        return _dynamo_item_to_return_vals(item)

    async def lookup_by_request_id_and_identity(
        self, request_id: str, user_identity: str
    ) -> ActionProviderPersistenceReturnType:
        filter_expression = Attr("request_id").eq(request_id) & Attr("creator").eq(
            user_identity
        )
        response = self.table.scan(FilterExpression=filter_expression)
        items = response.get("Items", (None,))
        if len(items) == 0:
            items = (None,)
        return _dynamo_item_to_return_vals(items[0])

    async def store_action(
        self,
        action: ActionStatus,
        request: Optional[ActionRequest] = None,
        creator_id: Optional[str] = None,
        extra_data: Optional[Dict[str, Any]] = None,
    ) -> ActionProviderPersistenceReturnType:
        action_json = json.dumps(action, cls=PersistenceJsonEncoder)

        item = {
            "action_id": action.action_id,
            "creator": action.creator_id,
            "action_status": action_json,
        }
        if request is not None:
            request_json = json.dumps(request, cls=PersistenceJsonEncoder)
            item["request_id"] = request.request_id
            item["action_request"] = request_json
        if extra_data is not None:
            item["extra_data"] = extra_data

        self.table.put_item(Item=item)
        return action, request, extra_data

    async def remove_action(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        del_resp = self.table.delete_item(
            Key={
                "action_id": action_id,
            },
            ReturnValues="ALL_OLD",
        )
        item = del_resp.get("Attributes")
        # Make it an iterable to pass in there
        return _dynamo_item_to_return_vals(item)



import datetime
import json
import logging
from contextlib import asynccontextmanager
from typing import Any, Dict, Optional, Set, Type, TypeVar

from globus_action_provider_tools import ActionRequest, ActionStatus, ActionStatusValue
from pydantic import BaseModel
from sqlmodel import Field, Session, SQLModel, create_engine, select

from .async_helpers import await_if_needed
from .persistence import (
    ActionProviderPersistence,
    ActionProviderPersistenceReturnType,
    PersistenceJsonEncoder,
)

log = logging.getLogger(__name__)

_action_property_conversions = {
    "status": lambda x: ActionStatusValue[x],
}

T = TypeVar("T", bound=BaseModel)


def _copy_into(
    src_model: BaseModel, dest_model: BaseModel, omits: Optional[Set[str]] = None
) -> BaseModel:
    for k, v in src_model:
        if omits and k not in omits:
            setattr(dest_model, k, v)
    return dest_model


def _json_to_object(json_val: Optional[str], model_class: Type[T]) -> Optional[T]:
    if json_val is None:
        return None
    props = json.loads(json_val)
    for val_to_convert, convert_fn in _action_property_conversions.items():
        model_val = props.get(val_to_convert)
        if model_val is not None:
            props[val_to_convert] = convert_fn(model_val)
    return model_class(**props)


class ActionTableModel(SQLModel, table=True):
    __tablename__ = "GlobusActions"
    action_id: str = Field(primary_key=True)
    request_id: str = Field(index=True)
    creator: str = Field(index=True)
    completion_time: Optional[datetime.datetime]
    action_status_json: str
    request_json: Optional[str]
    extra_data_json: Optional[str]

    def to_persistence_return_type(self) -> ActionProviderPersistenceReturnType:
        action_status = _json_to_object(self.action_status_json, ActionStatus)
        request = _json_to_object(self.request_json, ActionRequest)
        extra_data = _json_to_object(self.extra_data_json, dict)
        return (action_status, request, extra_data)


class SQLModelProviderPersistence(ActionProviderPersistence):
    def __init__(
        self,
        sessionmaker,
        use_async=False,
    ):
        self.sessionmaker = sessionmaker
        self.use_async = use_async

    @classmethod
    def create_engine(cls, db_url: str, create_tables=False, **kwargs):
        engine = create_engine(db_url, **kwargs)
        if create_tables:
            cls.create_tables(engine)
        return engine

    @classmethod
    def create_tables(cls, engine) -> None:
        SQLModel.metadata.create_all(engine)

    @classmethod
    async def async_create_engine(cls, db_url: str, create_tables=False, **kwargs):
        from sqlalchemy.ext.asyncio import create_async_engine

        engine = create_async_engine(db_url, future=True, **kwargs)
        if create_tables:
            await cls.async_create_tables(engine)
        return engine

    @classmethod
    async def async_create_tables(cls, engine) -> None:
        async with engine.begin() as conn:
            await conn.run_sync(SQLModel.metadata.create_all)

    @asynccontextmanager
    async def get_session(self, **kwargs) -> Session:
        session = self.sessionmaker(**kwargs)
        try:
            yield session
            await await_if_needed(session.commit())
        except Exception:
            await await_if_needed(session.rollback())
            raise
        finally:
            session.close()

    async def lookup_by_id(
        self, session: Session, model_class: Type, id_val: Any
    ) -> Any:
        model = session.get(model_class, id_val)
        if isinstance(model, BaseModel) or model is None:
            return model
        return await model

    async def lookup_by_action_id(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        async with self.get_session(expire_on_commit=False) as session:
            table_model = await self.lookup_by_id(session, ActionTableModel, action_id)
            if table_model is None:
                return (None, None, None)
            else:
                return table_model.to_persistence_return_type()

    async def lookup_by_request_id_and_identity(
        self, request_id: str, user_identity: str
    ) -> ActionProviderPersistenceReturnType:
        async with self.get_session(expire_on_commit=False) as session:
            table_model = (
                await await_if_needed(
                    session.exec(
                        select(ActionTableModel).where(
                            ActionTableModel.request_id == request_id
                            and ActionTableModel.creator == user_identity
                        )
                    )
                )
            ).first()
            if table_model is None:
                return (None, None, None)
            else:
                return table_model.to_persistence_return_type()

    async def store_action(
        self,
        action: ActionStatus,
        request: Optional[ActionRequest] = None,
        creator_id: Optional[str] = None,
        extra_data: Optional[Dict[str, Any]] = None,
    ) -> ActionProviderPersistenceReturnType:
        action_json = json.dumps(action, cls=PersistenceJsonEncoder)

        if request is not None:
            request_json = json.dumps(request, cls=PersistenceJsonEncoder)
            request_id = request.request_id
        else:
            request_json = None
            request_id = None
        if extra_data is not None:
            extra_data_json = json.dumps(extra_data, cls=PersistenceJsonEncoder)
        else:
            extra_data_json = None

        table_model = ActionTableModel(
            action_id=action.action_id,
            action_status_json=action_json,
            request_id=request_id,
            request_json=request_json,
            creator=creator_id,
            completion_time=action.completion_time,
            extra_data_json=extra_data_json,
        )

        rvals: ActionProviderPersistenceReturnType = (None, None, None)
        async with self.get_session() as session:
            current_entry: ActionTableModel = await self.lookup_by_id(
                session, ActionTableModel, action.action_id
            )
            if current_entry is not None:
                # The action, the completion time, and the extra data are the
                # only mutable fields
                current_entry.action_status_json = action_json
                current_entry.completion_time = action.completion_time
                current_entry.extra_data_json = extra_data_json
                session.add(current_entry)
                rvals = current_entry.to_persistence_return_type()
            else:
                session.add(table_model)
                rvals = table_model.to_persistence_return_type()
            session.commit()
        return rvals

    async def remove_action(
        self, action_id: str
    ) -> ActionProviderPersistenceReturnType:
        async with self.get_session(expire_on_commit=False) as session:
            table_model = await self.lookup_by_id(session, ActionTableModel, action_id)
            if table_model is not None:
                await await_if_needed(session.delete(table_model))

                return table_model.to_persistence_return_type()
            else:
                return (None, None, None)



from globus_action_provider_tools import (
    ActionProviderDescription,
    ActionRequest,
    ActionStatus,
    ActionStatusValue,
    AuthState,
)
from globus_action_provider_tools.data_types import (
    ActionFailedDetails,
    ActionInactiveDetails,
)

from .async_helpers import async_invoke, await_if_needed
from .fastapi_router import GlobusActionProviderRouter
from .in_memory_persistence import InMemoryActionProviderPersistence
from .persistence import ActionProviderPersistence, ActionProviderPersistenceReturnType

__all__ = (
    "async_invoke",
    "await_if_needed",
    "GlobusActionProviderRouter",
    "ActionProviderPersistence",
    "ActionProviderPersistenceReturnType",
    "ActionProviderDescription",
    "ActionRequest",
    "ActionStatus",
    "ActionStatusValue",
    "AuthState",
    "ActionFailedDetails",
    "ActionInactiveDetails",
    "InMemoryActionProviderPersistence",
)



import inspect
from typing import Any, Awaitable, Callable, Coroutine, Dict, List, TypeVar, Union

T = TypeVar("T")


async def async_invoke(
    func: Callable[[Any], Union[T, Coroutine[Any, Any, T]]], *args, **kwargs
) -> Awaitable[T]:
    sig = inspect.signature(func)
    call_args: List[Any] = []
    call_kwargs: Dict[str, Any] = {}
    args_index = 0
    for param_name, param_info in sig.parameters.items():
        if param_name in kwargs:
            call_kwargs[param_name] = kwargs[param_name]
        elif param_info.kind in {
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.POSITIONAL_ONLY,
        }:
            if args_index < len(args):
                call_args.append(args[args_index])
                args_index += 1
        elif param_info.kind is inspect.Parameter.VAR_KEYWORD:
            call_kwargs[param_name] = kwargs
        elif param_info.kind is inspect.Parameter.VAR_POSITIONAL:
            call_args.extend(args[args_index:])
            args_index = len(args)

    if inspect.iscoroutinefunction(func):
        r = await func(*call_args, **call_kwargs)
    else:
        r = func(*call_args, **call_kwargs)
    return r


async def await_if_needed(val: Union[T, Awaitable[T]]) -> T:
    if inspect.isawaitable(val):
        return await val
    else:
        return val



from typing import Any, Dict, Optional

from fastapi.exceptions import HTTPException


class ActionProviderException(HTTPException):
    def __init__(
        self,
        status_code: int,
        detail: Any = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(status_code, detail=detail, headers=headers)


class ActionNotFoundException(ActionProviderException):
    def __init__(self, action_id: str):
        super().__init__(
            status_code=404, detail=f"Action with id {action_id} not Found"
        )


class ActionConflictException(ActionProviderException):
    def __init__(self, action_id: str, detail: Optional[str]):
        super().__init__(status_code=409, detail=detail)


class ApplicationError(ActionProviderException):
    def __init__(self, detail: Optional[str]):
        super().__init__(status_code=500, detail=detail)


class ActionMethodMissingImplementation(ActionProviderException):
    def __init__(self, method_name: str):
        super().__init__(
            status_code=500,
            detail=f"Server failed to implement required method {method_name}",
        )


class UnauthorizedException(ActionProviderException):
    def __init__(self):
        super().__init__(
            status_code=401,
            detail=(
                "The server could not verify that you are authorized to access"
                " the URL requested. You either supplied the wrong credentials"
                " (e.g. a bad password), or your browser doesn't understand"
                " how to supply the credentials required."
            ),
        )
